#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.future.functools
# ---------------------------------------------------------------------
#
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
#
# Most of the code of this file is backported from Python 3.2 standard library
# with minor modifications to make it works on Python 2.7. So, this file is
# distributed under the terms of the Python Software Foundatation Licence for
# Python 3.2.
#
# Created on 2013-02-22
# Migrated to 'future' on 2016-09-16

'''Extensions to the `functools` module from the Python's standard library.

You may use this module as drop-in replacement of `functools`.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

from functools import *    # noqa
import functools as _stdlib    # noqa

from xoutil.deprecation import deprecate_linked
deprecate_linked()
del deprecate_linked

from xoutil.eight import python_version    # noqa
from xoutil.eight import callable    # noqa

import xoutil.fp.tools as fp
from xoutil.deprecation import deprecated


@deprecated(fp.pos_args)
class ctuple(fp.pos_args):
    '''Simple tuple marker for `compose`:func:.

    Since is a callable you may use it directly in ``compose`` instead of
    changing your functions to returns ctuples instead of tuples::

       >>> def compat_print(*args):
       ...     for arg in args:
       ...         print(arg)

       >>> compose(compat_print, ctuple, list, range, math=False)(3)
       0
       1
       2

       # Without ctuple prints the list
       >>> compose(compat_print, list, range, math=False)(10)
       [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    .. deprecated:: 1.8.0 Use `xoutil.fp.tool.pos_args`:class:.

    '''


# TODO: [med]  Should we port the `ctuple` to fp.tools?
@deprecated(fp.compose)
def compose(*callables, **kwargs):
    '''Returns a function that is the composition of several `callables`.

    By default `compose` behaves like mathematical function composition: this
    is to say that ``compose(f1, ... fn)`` is equivalent to ``lambda _x:
    fn(...(f1(_x))...)``.

    If any "intermediate" function returns a `ctuple`:class: it is expanded as
    several positional arguments to the next function.

    .. versionchanged:: 1.5.5 At least a callable must be passed, otherwise a
                        TypeError is raised.  If a single callable is passed
                        it is returned without change.

    :param math: Indicates if `compose` should behave like mathematical
                 function composition: last function in `funcs` is applied
                 last. If False, then the last function in `func` is applied
                 first.

    .. deprecated:: 1.8.0 Use `xoutil.fp.tools.compose`:class:.

    '''
    from xoutil.params import check_count
    check_count(callables, 1, caller='compose')
    if not all(callable(func) for func in callables):
        raise TypeError('Every func must a callable')
    if len(callables) == 1:
        return callables[0]
    math = kwargs.get('math', True)
    if not math:
        callables = list(reversed(callables))

    return fp.compose(*reversed(callables))


# TODO: Check relevance of the following function.
# The real signature should be (*funcs, times)
def power(*args):
    '''Returns the "power" composition of several functions.

    Examples::

       >>> import operator
       >>> f = power(partial(operator.mul, 3), 3)
       >>> f(23) == 3*(3*(3*23))
       True

       >>> power(operator.neg)
       Traceback (most recent call last):
       ...
       TypeError: power() takes at least 2 arguments (1 given)

    '''
    from xoutil.params import check_count
    from xoutil.fp.tools import compose
    check_count(args, 2, caller='power')
    funcs, times = args[:-1], args[-1]
    if any(not callable(func) for func in funcs):
        raise TypeError('Arguments of `power`, but last, must be callables')
    if not (isinstance(times, int) and times > 0):
        raise TypeError('Last argument of `power` must be a positive integer')
    if len(funcs) > 1:
        base = (compose(funcs), )
    else:
        base = (funcs[0], )
    return compose(*(base * times))


def lwraps(*args, **kwargs):
    '''Lambda wrapper.

    Useful for decorate lambda functions with name and documentation.

    As positional arguments could be passed the function to be decorated and
    the name in any order.  So the next two ``identity`` definitions are
    equivalents::

      >>> from xoutil.future.functools import lwraps as lw

      >>> identity = lw('identity', lambda arg: arg)

      >>> identity = lw(lambda arg: arg, 'identity')

    As keyword arguments could be passed some special values, and any number
    of literal values to be assigned:

    - **name**: The name of the function (``__name__``); only valid if not
      given as positional argument.

    - **doc**: The documentation (``__doc__`` field).

    - **wrapped**: An object to extract all values not yet assigned.  These
      values are ('__module__', '__name__' and '__doc__') to be assigned, and
      '__dict__' to be updated.

    If the function to decorate is present in the positional arguments, this
    same argument function is directly returned after decorated; if not a
    decorator is returned similar to standard `wraps`:func:.

    For example::

      >>> from xoutil.future.functools import lwraps as lw

      >>> is_valid_age = lw('is-valid-human-age', lambda age: 0 < age <= 120,
      ...                   doc=('A predicate to evaluate if an age is '
      ...                        'valid for a human being.')

      >>> @lw(wrapped=is_valid_age)
      ... def is_valid_working_age(age):
      ...     return 18 < age <= 70

      >>> is_valid_age(16)
      True

      >>> is_valid_age(200)
      False

      >>> is_valid_working_age(16)
      False

    .. versionadded:: 1.7.0

    '''
    from types import FunctionType, MethodType
    from xoutil.symbols import Unset
    from xoutil.eight import string_types, iteritems
    from xoutil.eight import string
    from xoutil.params import check_count

    def repeated(name):
        msg = "lwraps got multiple values for argument '{}'"
        raise TypeError(msg.format(name))

    def settle_str(name, value):
        if value is not Unset:
            if isinstance(value, string_types):
                if name not in source:
                    source[name] = value
                else:
                    repeated(name)
            else:
                from xoutil.eight import type_name
                msg = 'lwraps expecting string for "{}", {} found'
                raise TypeError(msg.format(name, type_name(value)))

    methods = (staticmethod, classmethod, MethodType)
    decorables = methods + (FunctionType, )

    name_key = '__name__'
    doc_key = '__doc__'
    mod_key = '__module__'
    safes = {name_key, mod_key}
    source = {}
    target = Unset
    count = len(args)
    check_count(count, 0, 2, caller='lwraps')
    i = 0
    while i < count:
        arg = args[i]
        if isinstance(arg, string_types):
            settle_str(name_key, arg)
        elif isinstance(arg, decorables):
            if target is Unset:
                target = arg
            else:
                repeated('target-function')
        else:
            msg = 'lwraps arg {} must be a string or decorable function'
            raise TypeError(msg.format(i))
        i += 1
    wrapped = kwargs.pop('wrapped', Unset)
    settle_str(name_key, kwargs.pop('name', Unset))
    settle_str(name_key, kwargs.pop(name_key, Unset))
    settle_str(doc_key, kwargs.pop('doc', Unset))
    settle_str(doc_key, kwargs.pop(doc_key, Unset))
    source.update(kwargs)
    if wrapped is not Unset:
        # TODO: Check the type of `wrapped` to find these attributes in
        # disparate callable objects similarly with functions.
        for name in (mod_key, name_key, doc_key):
            if name not in source:
                source[str(name)] = getattr(wrapped, name)
        d = source.setdefault('__dict__', {})
        d.update(wrapped.__dict__)

    def wrapper(target):
        if isinstance(target, decorables):
            res = target
            if isinstance(target, methods):
                target = target.__func__
            for name in (mod_key, name_key, doc_key):
                if name in source:
                    value = source.pop(name)
                    if name in safes:
                        value = string.force(value)
                    setattr(target, str(name), value)
                d = source.pop('__dict__', Unset)
                if d:
                    target.__dict__.update(d)
            for key, value in iteritems(source):
                setattr(target, key, value)
            return res
        else:
            from xoutil.eight import type_name
            msg = 'only functions are decorated, not {}'
            raise TypeError(msg.format(type_name(target)))

    return wrapper(target) if target else wrapper

    # TODO: Next code could be removed.
    # func.__name__ = string.force(name)
    # if doc:
    #     func.__doc__ = doc
    # return func


try:
    from functools import _CacheInfo
except ImportError:
    from xoutil.future.collections import namedtuple

    _CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize",
                                          "currsize"])

if python_version < 3.3:
    from threading import RLock
    # Back-ported lru_cache from py33. But take note that if running with at
    # least py3 we will use Python's version, so don't mess with internals.

    WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__qualname__', '__doc__',
                           '__annotations__')
    WRAPPER_UPDATES = ('__dict__',)

    def update_wrapper(wrapper,
                       wrapped,
                       assigned=WRAPPER_ASSIGNMENTS,
                       updated=WRAPPER_UPDATES):
        """Update a wrapper function to look like the wrapped function

           wrapper is the function to be updated
           wrapped is the original function
           assigned is a tuple naming the attributes assigned directly
           from the wrapped function to the wrapper function (defaults to
           functools.WRAPPER_ASSIGNMENTS)
           updated is a tuple naming the attributes of the wrapper that
           are updated with the corresponding attribute from the wrapped
           function (defaults to functools.WRAPPER_UPDATES)
        """
        wrapper.__wrapped__ = wrapped
        for attr in assigned:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
        for attr in updated:
            getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
        # Return the wrapper so this can be used as a decorator via partial()
        return wrapper

    def wraps(wrapped,
              assigned=WRAPPER_ASSIGNMENTS,
              updated=WRAPPER_UPDATES):
        """Decorator factory to apply update_wrapper() to a wrapper function

           Returns a decorator that invokes update_wrapper() with the decorated
           function as the wrapper argument and the arguments to wraps() as the
           remaining arguments. Default arguments are as for update_wrapper().
           This is a convenience function to simplify applying partial() to
           update_wrapper().
        """
        return partial(update_wrapper, wrapped=wrapped,    # noqa
                       assigned=assigned, updated=updated)

    class _HashedSeq(list):
        """ This class guarantees that hash() will be called no more than once
            per element.  This is important because the lru_cache() will hash
            the key multiple times on a cache miss.

        """

        __slots__ = 'hashvalue'

        def __init__(self, tup, hash=hash):
            self[:] = tup
            self.hashvalue = hash(tup)

        def __hash__(self):
            return self.hashvalue

    def _make_key(args, kwds, typed,
                  kwd_mark=(object(),),
                  fasttypes={int, str, frozenset, type(None)},
                  sorted=sorted, tuple=tuple, type=type, len=len):
        """Make a cache key from optionally typed positional and keyword
        arguments.

        The key is constructed in a way that is flat as possible rather than
        as a nested structure that would take more memory.

        If there is only a single argument and its data type is known to cache
        its hash value, then that argument is returned without a wrapper.
        This saves space and improves lookup speed.

        """
        key = args
        if kwds:
            sorted_items = sorted(kwds.items())
            key += kwd_mark
            for item in sorted_items:
                key += item
        if typed:
            key += tuple(type(v) for v in args)
            if kwds:
                key += tuple(type(v) for k, v in sorted_items)
        elif len(key) == 1 and type(key[0]) in fasttypes:
            return key[0]
        return _HashedSeq(key)

    def lru_cache(maxsize=128, typed=False):
        """Decorator to wrap a function with a memoizing callable that saves
        up to the `maxsize` most recent calls.  It can save time when an
        expensive or I/O bound function is periodically called with the same
        arguments.

        Since a dictionary is used to cache results, the positional and
        keyword arguments to the function must be hashable.

        If `maxsize` is set to None, the LRU feature is disabled and the cache
        can grow without bound.  The LRU feature performs best when maxsize is
        a power-of-two.

        If `typed` is set to True, function arguments of different types will
        be cached separately.  For example, ``f(3)`` and ``f(3.0)`` will be
        treated as distinct calls with distinct results.

        To help measure the effectiveness of the cache and tune the `maxsize`
        parameter, the wrapped function is instrumented with a cache_info()
        function that returns a named tuple showing hits, misses, maxsize and
        currsize.  In a multi-threaded environment, the hits and misses are
        approximate.

        The decorator also provides a ``cache_clear()`` function for clearing
        or invalidating the cache.

        The original underlying function is accessible through the
        ``__wrapped__`` attribute.  This is useful for introspection, for
        bypassing the cache, or for rewrapping the function with a different
        cache.

        An `LRU (least recently used)`__ cache works best when the most recent
        calls are the best predictors of upcoming calls (for example, the most
        popular articles on a news server tend to change each day).  The
        cache's size limit assures that the cache does not grow without bound
        on long-running processes such as web servers.

        __ http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used

        """

        # Users should only access the lru_cache through its public API:
        #       cache_info, cache_clear, and f.__wrapped__

        # The internals of the lru_cache are encapsulated for thread safety
        # and to allow the implementation to change (including a possible C
        # version).

        # Constants shared by all lru cache instances:
        sentinel = object()        # unique object used to signal cache misses
        make_key = _make_key       # build a key from the function arguments
        PREV, NEXT, KEY, RESULT = 0, 1, 2, 3   # names for the link fields

        def decorating_function(user_function):
            cache = {}
            hits = misses = 0
            _cache_vars = [hits, misses]  # hits, misses
            _HITS, _MISSES = 0, 1
            full = False
            # bound method to lookup a key or return None
            cache_get = cache.get
            # because linkedlist updates aren't threadsafe
            lock = RLock()
            # root of the circular doubly linked list
            root = []
            # initialize by pointing to self
            root[:] = [root, root, None, None]
            _cache_vars.extend([full, root])
            _FULL = 2
            _ROOT = 3

            if maxsize == 0:

                def wrapper(*args, **kwds):
                    # No caching -- just a statistics update after a
                    # successful call
                    # nonlocal misses
                    result = user_function(*args, **kwds)
                    _cache_vars[_MISSES] += 1
                    return result

            elif maxsize is None:

                def wrapper(*args, **kwds):
                    # Simple caching without ordering or size limit nonlocal
                    # hits, misses
                    key = make_key(args, kwds, typed)
                    result = cache_get(key, sentinel)
                    if result is not sentinel:
                        _cache_vars[_HITS] += 1
                        return result
                    result = user_function(*args, **kwds)
                    cache[key] = result
                    _cache_vars[_MISSES] += 1
                    return result

            else:

                def wrapper(*args, **kwds):
                    # Size limited caching that tracks accesses by recency
                    # nonlocal root, hits, misses, full
                    root = _cache_vars[_ROOT]
                    key = make_key(args, kwds, typed)
                    with lock:
                        link = cache_get(key)
                        if link is not None:
                            # Move the link to the front of the circular queue
                            link_prev, link_next, _key, result = link
                            link_prev[NEXT] = link_next
                            link_next[PREV] = link_prev
                            last = root[PREV]
                            last[NEXT] = root[PREV] = link
                            link[PREV] = last
                            link[NEXT] = root
                            _cache_vars[_HITS] += 1
                            return result
                    result = user_function(*args, **kwds)
                    with lock:
                        if key in cache:
                            # Getting here means that this same key was added
                            # to the cache while the lock was released.  Since
                            # the link update is already done, we need only
                            # return the computed result and update the count
                            # of misses.
                            pass
                        elif _cache_vars[_FULL]:
                            # Use the old root to store the new key and result.
                            oldroot = root = _cache_vars[_ROOT]
                            oldroot[KEY] = key
                            oldroot[RESULT] = result
                            # Empty the oldest link and make it the new root.
                            # Keep a reference to the old key and old result
                            # to prevent their ref counts from going to zero
                            # during the update. That will prevent potentially
                            # arbitrary object clean-up code (i.e. __del__)
                            # from running while we're still adjusting the
                            # links.
                            root = _cache_vars[_ROOT] = oldroot[NEXT]
                            oldkey = root[KEY]
                            # TODO: never used
                            oldresult = root[RESULT]    # noqa
                            root[KEY] = root[RESULT] = None
                            # Now update the cache dictionary.
                            del cache[oldkey]
                            # Save the potentially reentrant cache[key]
                            # assignment for last, after the root and links
                            # have been put in a consistent state.
                            cache[key] = oldroot
                        else:
                            # Put result in a new link at the front of the
                            # queue.
                            last = root[PREV]
                            link = [last, root, key, result]
                            last[NEXT] = root[PREV] = cache[key] = link
                            _cache_vars[_FULL] = (len(cache) >= maxsize)
                        _cache_vars[_MISSES] += 1
                    return result

            def cache_info():
                """Report cache statistics"""
                with lock:
                    hits, misses = _cache_vars[:2]
                    return _CacheInfo(hits, misses, maxsize, len(cache))

            def cache_clear():
                """Clear the cache and cache statistics"""
                # nonlocal hits, misses, full
                with lock:
                    hits, misses, full, root = _cache_vars
                    cache.clear()
                    root[:] = [root, root, None, None]
                    hits = misses = 0
                    full = False
                    _cache_vars[:] = [hits, misses, full, root]

            wrapper.cache_info = cache_info
            wrapper.cache_clear = cache_clear
            return update_wrapper(wrapper, user_function)

        return decorating_function


def curry(f):
    '''Return a function that automatically 'curries' is positional arguments.

    Example::

        >>> add = curry(lambda x, y: x + y)
        >>> add(1)(2)
        3

        >>> add(1, 2)
        3

        >>> add()()()(1, 2)
        3
    '''
    from xoutil.future.inspect import getfullargspec
    fargs = getfullargspec(f)[0]

    def curried(cargs=None):
        if cargs is None:
            cargs = []

        def inner(*args, **kwargs):
            cargs_ = cargs + list(args)
            if len(cargs_) < len(fargs):
                return curried(cargs_)
            else:
                return f(*cargs_, **kwargs)
        return inner
    return curried()



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.future.types
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# Copyright (c) 2010-2012 Medardo Rodriguez
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Migrated to 'future' on 2016-09-12

'''Extends the standard `types` module.

Standard module defines names for all type symbols known in the standard
interpreter.

This modules mirrors all the functions (and, in general, objects) from the
standard library module `types`:mod:; but it also includes several new types
and type-related functions.

This module unifies old `xoutil.types`:mod: and `xoutil.eigth.types`:mod:
modules, which are deprecated now.

There are some symbols included in Python 2.x series, but not in Python 3 that
we don't include here:

- `TypeType`: use `type` instead

- `ObjectType`: use `object`

- `IntType`: use `int`

- `LongType`: use `long` in Python 2 and `int` in Python 3; better see
  `xoutil.eight.integer_types` definition.

- `FloatType`: use `float`

- `BooleanType`: use `bool`

- `ComplexType`: use `complex`

- `StringType`: use str

- `UnicodeType`: use `unicode` in Python 2 and `str` in Python 3; there are
  two aliases for that: `xoutil.eigth.UnicodeType` and
  `xoutil.eigth.text_type`.

- `StringTypes`: use `xoutil.eigth.StringTypes`or
  `xoutil.eigth.string_types`.

- `BufferType`: use `buffer` in Python 2 and `memoryview` in Python 3; there
  is an alias for this convention in `xoutil.eigth.buffer`.  The
  `memoryview`:class: API is similar but not exactly the same as that of
  `buffer`.

- `TupleType`: use `tuple`

- `ListType`: use `list`

- `DictType` (or `DictionaryType`): use `dict`

- `ClassType`: Python 2 old-style classes, don't exists in Python 3, see
  `xoutil.eigth.ClassTypes`.

- `InstanceType`: type of instances of Python 2 old-style classes, don't
  exists in Python 3, see `xoutil.eigth.typeof`.

- `UnboundMethodType`: use `~types.MethodType` alias

- `FileType`: use `file`

- `XRangeType` (or `xrange`): in Python 3 `range` always returns a generator
  object; use `xoutil.eigth.range`:class: for compatibility; wraps it with
  list (``list(range(...))``) to obtain old `range` style

- `SliceType`: use `slice`

- In Jython and PyPy, `MemberDescriptorType` is identical to
  `GetSetDescriptorType`; to mantain compatibility in some `xoutil` code, they
  are differentiated in this module.

- `CoroutineType` and `coroutine`: use new Python 3 `async` statement; not
  implementable in Python 2.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from types import *    # noqa
import types as _stdlib    # noqa

from xoutil.deprecation import deprecate_linked
deprecate_linked()
del deprecate_linked

from xoutil.eight import python_version    # noqa

try:
    from types import __all__    # noqa
    __all__ = list(__all__)
except ImportError:
    # Python 3.3 don't implement '__all__' for 'string' module.
    __all__ = [name for name in dir(_stdlib) if not name.startswith('_')]

try:
    NoneType    # noqa
except NameError:
    NoneType = type(None)
    __all__.append('NoneType')

try:
    EllipsisType    # noqa
except NameError:
    EllipsisType = type(Ellipsis)
    __all__.append('EllipsisType')

try:
    DictProxyType    # noqa
except NameError:
    DictProxyType = type(type.__dict__)
    __all__.append('DictProxyType')

try:
    MappingProxyType    # noqa
except NameError:
    from xoutil.eight._types import MappingProxyType
    if MappingProxyType is not DictProxyType:
        MappingProxyType.register(DictProxyType)
    __all__.append('MappingProxyType')

if python_version == 2:
    from collections import Mapping
    if not issubclass(MappingProxyType, Mapping):
        # TODO: when implement `xoutil.future.collections`, fix this problem
        # there.
        Mapping.register(MappingProxyType)
    del Mapping

try:
    NotImplementedType    # noqa
except NameError:
    NotImplementedType = type(NotImplemented)
    __all__.append('NotImplementedType')


# Check Jython and PyPy peculiarity
if MemberDescriptorType is GetSetDescriptorType:    # noqa
    class _foo(object):
        __slots__ = 'bar'
    MemberDescriptorType = type(_foo.bar)
    del _foo

FuncTypes = tuple({FunctionType, MethodType, LambdaType,    # noqa
                   BuiltinFunctionType, BuiltinMethodType})    # noqa

from xoutil.eight import class_types    # noqa
func_types = FuncTypes    # Just an alias

# These types are defined in `inspect` module for Python >= 3.3
MethodWrapperType = type(all.__call__)
WrapperDescriptorType = type(type.__call__)    # In PyPy is MethodWrapperType
ClassMethodWrapperType = type(dict.__dict__['fromkeys'])


sn_ok = python_version >= 3.4
if sn_ok:
    try:
        SimpleNamespace    # noqa
    except NameError:
        sn_ok = False
        __all__.append('SimpleNamespace')

if not sn_ok:
    from abc import ABCMeta
    from xoutil.eight.meta import metaclass

    from xoutil.reprlib import recursive_repr

    class SimpleNamespace(metaclass(ABCMeta)):
        '''A simple attribute-based namespace.

        SimpleNamespace(**kwargs)

        '''

        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def __eq__(self, other):
            # TODO: This method is not implemented in py33
            ok = isinstance(other, SimpleNamespace)
            return ok and self.__dict__ == other.__dict__

        @recursive_repr(str('namespace(...)'))
        def __repr__(self):
            keys = sorted(self.__dict__)
            items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
            return "{}({})".format('namespace', ", ".join(items))

    del recursive_repr, ABCMeta, metaclass

    try:
        from types import SimpleNamespace as _sns
        # TODO: @manu, why needed in this case?
        SimpleNamespace.register(_sns)
        del _sns
    except ImportError:
        pass

try:
    DynamicClassAttribute    # noqa
except NameError:
    # TODO: Add tests
    class DynamicClassAttribute(property):
        '''Route attribute access on a class to `~object.__getattr__`:meth:.

        This is a descriptor, used to define attributes that act differently
        when accessed through an instance and through a class.  Instance
        access remains normal, but access to an attribute through a class will
        be routed to the class's `~object.__getattr__`:meth: method;
        this is done by raising `AttributeError`:class:.

        This allows one to have properties active on an instance, and have
        virtual attributes on the class with the same name (see
        `~py3:enum.Enum`:class: for an example).

        .. versionadded:: 1.5.5

        .. versionchanged:: 1.8.0 Inherits from `property`

        .. note:: The class `Enum` mentioned has not yet been back-ported.

        .. note:: In Python version>=3.4 this is an alias to
                  `types.DynamicClassAttribute
                  <py3:types.DynamicClassAttribute>`:class:.

        '''
        def __init__(self, fget=None, fset=None, fdel=None, doc=None):
            super(DynamicClassAttribute, self).__init__(fget, fset, fdel, doc)
            # support for abstract methods in Python 2
            isabs = bool(getattr(fget, '__isabstractmethod__', False))
            self.__isabstractmethod__ = isabs

        def __get__(self, obj, owner=None):
            if obj is None:
                if self.__isabstractmethod__:
                    return self
                else:
                    raise AttributeError()
            else:
                return super(DynamicClassAttribute, self).__get__(obj, owner)

    __all__.append('DynamicClassAttribute')

try:
    new_class    # noqa
except NameError:
    from xoutil.eight._types import new_class    # noqa
    __all__.append('new_class')

try:
    prepare_class    # noqa
except NameError:
    from xoutil.eight._types import prepare_class    # noqa
    __all__.append('prepare_class')

try:
    from types import _calculate_meta
except ImportError:
    from xoutil.eight._types import _calculate_meta    # noqa

if __name__ == 'xoutil.types':
    from xoutil.deprecation import deprecated
    from xoutil.symbols import Unset as _unset
    from collections import Mapping
    from xoutil.eight import force_type as type_coerce

    # TODO: @manu, consider for future types
    from re import compile as _regex_compile
    RegexPattern = type(_regex_compile(''))
    del _regex_compile

    # TODO: used internally in this module
    class mro_dict(Mapping):
        '''Utility behaving like a read-only dict of `target` MRO attributes.

        For example::

          >>> class A(object):
          ...     x = 12
          ...     y = 34

          >>> class B(A):
          ...     y = 56
          ...     z = 78

          >>> d = mro_dict(B)

          >>> d['x']
          12

          >>> d['y']
          56

          >>> d['z']
          78

        '''
        # TODO: What is the application for this utility?
        __slots__ = ('_probes', '_keys')

        def __init__(self, target):
            from xoutil.future.inspect import _static_getmro
            type_ = type_coerce(target)
            target_mro = _static_getmro(type_)
            self._probes = tuple(c.__dict__ for c in target_mro)
            self._keys = set()

        def __getitem__(self, name):
            from xoutil.objects import get_first_of
            result = get_first_of(self._probes, name, default=_unset)
            if result is not _unset:
                return result
            else:
                raise KeyError(name)

        def __iter__(self):
            if not self._keys:
                self._settle_keys()
            return iter(self._keys)

        def __len__(self):
            if not self._keys:
                self._settle_keys()
            return len(self._keys)

        def _settle_keys(self):
            for probe in self._probes:
                for key in probe:
                    if key not in self._keys:
                        self._keys.add(key)

    # TODO: used internally in this module
    def mro_get_value_list(cls, name):
        '''Return a list with all `cls` class attributes in MRO.'''
        from xoutil.future.inspect import _static_getmro
        mro = _static_getmro(type_coerce(cls))
        return [t.__dict__[name] for t in mro if name in t.__dict__]

    # TODO: not further use
    def mro_get_full_mapping(cls, name):
        '''Return a dictionary with all items from `cls` in MRO.

        All values corresponding to `name` must be valid mappings.

        '''
        aux = mro_get_value_list(cls, name)
        count = len(aux)
        if count == 0:
            return {}
        elif count == 1:
            return aux[0]
        else:
            res = {}
            for m in aux:
                for key in m:
                    if key not in res:
                        res[key] = m[key]
            return res

    # TODO: Many of is_*method methods here are needed to be compared against
    # the standard lib's module inspect versions.  If they behave the same,
    # these should be deprecated in favor of the standards.

    # TODO: use in `xoutil.json.JSONEncoder` was replaced by
    # ``isinstance(maybe, Iterable)``, so now is used only internally.
    def is_iterable(maybe):
        '''Returns True if `maybe` is an iterable object (e.g. implements the
        `__iter__` method):

        ::

            >>> is_iterable('all strings are iterable')
            True

            # Numbers are not
            >>> is_iterable(1)
            False

            >>> from xoutil.eight import range
            >>> is_iterable(range(1))
            True

            >>> is_iterable({})
            True

            >>> is_iterable(tuple())
            True

            >>> is_iterable(set())
            True

        '''
        try:
            iter(maybe)
        except:
            return False
        else:
            return True

    # TODO: @manu, external uses to be upgraded:
    #
    # - `odoo.addons.mail.xopgi.index`: can be replaced by a local function
    #   ``isinstance(maybe, (tuple, list, set))`` or by
    #   `xoutil.values.simple.logic_iterable_coerce`.
    #
    # - `xopgi.account.xopgi.xopgi_proper_currency.move`: ibidem
    #
    # - `xopgi.base.xopgi.xopgi_recurrence.models.recurrent_model`: two uses,
    #   the first is similar to above examples, second use the concept of
    #   arrays (mappings are collections but not arrays) and maybe must
    #   consider to move this function to new `xoutil.future.collections` and
    #   use it from there.  @med, consider
    #   `xoutil.values.simple.array_coerce`.
    #
    # - `xopgi.hr.xopgi.xopgi_hr_job_assets.hr_job`: ibidem to first.
    #
    # - `xopgi.hr_contract.xopgi.xopgi_hr_contract.hr_contract`: ibidem.
    #
    # - `xequiotl.rstmodel`: ibidem.
    #
    # - `xoeuf.cli.mailgate`: ibidem.
    #
    # - `xoeuf.cli.secure`: ibidem.
    #
    # - `xoeuf.osv.model_extensions`: ibidem.
    #
    # - `xoeuf.osv.writers`: ibidem.
    #
    # - `xotl.simple.event`: ibidem.
    #
    # - `xoutil.tests.test_types`: reconfigure all tests after upgrading all
    #   uses.
    #
    # - `xoutil.cpystack`: migrated to use new
    #   `xoutil.values.simple.force_sequence_coerce`.
    #
    # - `xoutil.fs`: migrated to use
    #    `xoutil.values.simple.force_iterable_coerce`.
    #
    # - `xoutil.iterators`: replaced by ``isinstance(fill, Iterable)``.
    #
    # - `xoutil.objects`: replaced by
    #   `xoutil.values.simple.logic_collection_coerce` or
    #   `xoutil.values.simple.logic_iterable_coerce`.
    def is_collection(maybe):
        '''Test `maybe` to see if it is a tuple, a list, a set or a generator
        function.

        It returns False for dictionaries and strings::

            >>> is_collection('all strings are iterable')
            False

            # Numbers are not
            >>> is_collection(1)
            False

            >>> from xoutil.eight import range
            >>> is_collection(range(1))
            True

            >>> is_collection({})
            False

            >>> is_collection(tuple())
            True

            >>> is_collection(set())
            True

            >>> is_collection(a for a in range(100))
            True

        .. versionchanged:: 1.5.5 UserList are collections.

        '''
        from xoutil.values.simple import logic_collection_coerce, nil
        return logic_collection_coerce(maybe) is not nil

    # TODO: There was only one reference to this in `xoutil.objects`, see how
    # was replaced in the new body of this function
    def is_mapping(maybe):
        '''Test `maybe` to see if it is a valid mapping.'''
        from xoutil.future.collections import Mapping
        return isinstance(maybe, Mapping)

    # only referenced locally by `is_scalar`.
    def is_string_like(maybe):
        '''Returns True if `maybe` acts like a string'''
        try:
            maybe + ""
        except TypeError:
            return False
        else:
            return True

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xoonko.core.management.profiles.loaders.models_loader`
    #
    # - `xoonko.core.management.profiles.producers.json_producer`
    #
    # - `xoutil.iterators`
    def is_scalar(maybe):
        '''Returns True if `maybe` is a string, an int, or some other scalar type
        (i.e not an iterable.)

        '''
        return is_string_like(maybe) or not is_iterable(maybe)

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xoutil.objects`: only a reference in a comment, must be removed.
    #
    # - `xoutil.records`: fixed.
    def is_staticmethod(desc, name=_unset):
        '''Returns true if a `method` is a static method.

        This function takes the same arguments as `is_classmethod`:func:.

        '''
        if name:
            desc = mro_dict(desc).get(name, None)
        return isinstance(desc, staticmethod)

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xoutil.objects`: only a reference in a comment, must be removed.
    def is_classmethod(desc, name=_unset):
        '''Returns true if a `method` is a class method.

        :param desc: This may be the method descriptor or the class that holds
               the method, in the second case you must provide the `name` of
               the method.

               .. note::

                  Notice that in the first case what is needed is the **method
                  descriptor**, i.e, taken from the class' `__dict__`
                  attribute. If instead you pass something like
                  ``cls.methodname``, this method will return False whilst
                  `is_instancemethod`:func: will return True.

        :param name: The name of the method, if the first argument is the
               class.

        '''
        if name:
            desc = mro_dict(desc).get(name, None)
        return isinstance(desc, classmethod)

    # not used outside this module.
    def is_instancemethod(desc, name=_unset):
        '''Returns true if a given `method` is neither a static method nor a class
        method.

        This function takes the same arguments as `is_classmethod`:func:.

        '''
        if name:
            desc = mro_dict(desc).get(name, None)
        return isinstance(desc, FunctionType)    # noqa

    # not used outside this module.
    def is_module(maybe):
        '''Returns True if `maybe` is a module.'''
        return isinstance(maybe, ModuleType)    # noqa

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xoutil.objects.smart_copy`.
    @deprecated(type(_unset), 'This is only used in `~objects.smart_copy`.')
    class Required(object):
        '''A type for required fields in scenarios where a default is not
        possible.

        '''
        def __init__(self, *args, **kwargs):
            pass

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xopgi_recurrence.models.recurrent_model`: sentence
    #   ``are_instances(leftres, rightres, tuple)`` must be replaced by
    #   ``isinstance(leftres, tuple) and isinstance(rightres, tuple)``.
    #
    # - `xoutil.objects`
    #
    # Real "Py4k" signature ``are_instances(*subjects, types)``.
    def are_instances(*args):
        '''Return True if every `subject` is an instance of (any) `types`.

        :param subjects: All but last positional arguments.  Are the objects
            required to be instances of `types`.

        :param types: The last positional argument.  Either a single type or a
           sequence of types.  This must meet the conditions on the last
           argument of `isinstance`:func:.

        :returns: True or False.  True if for every `subject`,
           ``isinstance(subject, types)`` is True.  Otherwise, False.

        If no `subjects` are provided return True::

            >>> are_instances(int)
            True

        .. seealso:: The function `no_instances`:func: allows to test for
                     subjects not being instances of types.

        '''
        from xoutil.params import check_count
        check_count(args, 1, caller='are_instances')
        subjects, types = args[:-1], args[-1]
        if not subjects:
            isinstance(None, types)   # HACK: always validate `types`.
        return all(isinstance(subject, types) for subject in subjects)

    # TODO: @manu, @med, review external references to this function, and
    # remove them:
    #
    # - `xoutil.objects`
    #
    # Real Py4k signature ``are_instances(*subjects, types)``self.
    def no_instances(*args):
        '''Return True if every `subject` is **not** an instance of (neither)
        `types`.

        :param subjects: All but last positional arguments.  Are the objects
               required not to be instances of `types`.

        :param types: The last positional argument.  Either a single type or a
               sequence of types.  This must meet the conditions on the last
               argument of `isinstance`:func:.

        :returns: True or False.  True if for every `subject`,
                  ``isinstance(subject, types)`` is False.  Otherwise, False.

        If no `subjects` are provided return True::

            >>> no_instances(int)
            True

        .. note:: This is not the same as ``not are_instances(...)``.

           This function requires that *no* subject is an instance of `types`.
           Negating `are_instances`:func: would be True if *any* subject is
           not an instance of `types`.

        '''
        from xoutil.params import check_count
        check_count(args, 1, caller='no_instances')
        subjects, types = args[:-1], args[-1]
        if not subjects:
            isinstance(None, types)   # HACK: always validate `types`.
        return all(not isinstance(subject, types) for subject in subjects)

    del deprecated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.future.calendar
# ---------------------------------------------------------------------
# Copyright (c) 2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-06-29

'''Extends the standard `calendar` module.

Xoutil extensions:

- Settle constants for all months.

Original documentation:

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from calendar import *    # noqa
import calendar as _stdlib    # noqa

__doc__ += _stdlib.__doc__

try:
    __all__ = list(_stdlib.__all__)
except AttributeError:
    pass


def _get_month_name_list():
    import sys
    import locale
    try:
        locale.setlocale(locale.LC_TIME, ('en_US', 'UTF-8'))
        mod = sys.modules[__name__]
        month = 1
        while month <= 12:
            name = _stdlib.month_name[month]
            if not hasattr(mod, name):
                setattr(mod, name, month)
                month += 1
    finally:
        locale.resetlocale(locale.LC_TIME)



_get_month_name_list()

del _get_month_name_list



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.queue
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-11-19

'''Standardization of Queue Processing.

In Python 2, the Queue processing module (A multi-producer, multi-consumer
queue) has the name with initial capital letter, nevertheless in Python 3 the
name is all in lower-case.

This is a compatibility module, importing symbols from this will be the same
in both versions of Python.

In Python 3 module there are two symbols imported from ``heapq`` module, those
are imported here for full compatibility.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

try:
    from Queue import (Queue, LifoQueue, PriorityQueue, Empty, Full)    # noqa
except ImportError:
    from queue import (Queue, LifoQueue, PriorityQueue, Empty, Full)    # noqa

try:
    from heapq import (heappop, heappush)    # noqa
except ImportError:
    from queue import (heappop, heappush)    # noqa

# best available implementation of current time in seconds
import time as _time

time = getattr(_time, 'monotonic', _time.time)

del _time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Technical string handling.

Technical strings are those that requires to be instances of `str` standard
type.  See `py-string-ambiguity`:any: for more information.

This module will be used mostly as a namespace, for example::

  from xoutil.eight import string
  Foobar.__name__ = string.force(class_name)

If these functions are going to be used standalone, do something like::

  from xoutil.eight.string import force as force_str
  Foobar.__name__ = force_str(class_name)

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_import)


def force(value=str()):
    '''Convert any value to standard `str` type in a safe way.

    This function is useful in some scenarios that require `str` type (for
    example attribute ``__name__`` in functions and types).

    As ``str is bytes`` in Python 2, using str(value) assures correct these
    scenarios in most cases, but in other is not enough, for example::

      >>> from xoutil.eight import string
      >>> def inverted_partial(func, *args, **keywords):
      ...     def inner(*a, **kw):
      ...         a += args
      ...         kw.update(keywords)
      ...         return func(*a, **kw)
      ...     name = func.__name__.replace('lambda', u'λ')
      ...     inner.__name__ = string.force(name)
      ...     return inner

    '''
    from xoutil.future.codecs import safe_decode, safe_encode
    if isinstance(value, str):
        return value
    elif str is bytes:      # Python 2
        return safe_encode(value)
    else:
        return safe_decode(value)


def safe_join(separator, iterable):
    '''Similar to `join` method in string objects.

    The semantics is equivalent to ``separator.join(iterable)`` but forcing
    separator and items to be of ``str`` standard type.

    For example::

      >>> safe_join('-', range(6))
      '0-1-2-3-4-5'

    Check that the expression ``'-'.join(range(6))`` raises a ``TypeError``.

    '''
    return force(separator).join(force(item) for item in iterable)


def force_ascii(value):
    '''Return the string normal form for the `value`

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    import unicodedata
    from xoutil.future.codecs import safe_decode
    from xoutil.eight import text_type
    from xoutil.eight import string
    if not isinstance(value, text_type):
        value = safe_decode(value)
    res = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    return string.force(res)


# ------------------ Here, the original file starts ------------------

if hasattr(str, 'isidentifier'):
    def isidentifier(s):
        return str(s) if s.isidentifier() else False
else:
    import re    # noqa
    _PY2_IDENTIFIER_REGEX = re.compile('(?i)^[_a-z][_a-z0-9]*$')
    del re

    def isidentifier(s):
        return str(s) if _PY2_IDENTIFIER_REGEX.match(s) else False

isidentifier.__doc__ = ('If `s` is a valid identifier according to the '
                        'language definition.')


def isfullidentifier(s):
    '''Check if `arg` is a valid dotted Python identifier.

    See `isidentifier`:func: for what "validity" means.

    '''
    return str(s) if all(isidentifier(p) for p in s.split('.')) else False


def safe_isidentifier(s):
    '''If `s` is a valid identifier according to the language definition.

    Check before if `s` is instance of string types.

    '''
    from xoutil.eight import string_types
    return isinstance(s, string_types) and isidentifier(s)


def safe_isfullidentifier(s):
    '''Check if `arg` is a valid dotted Python identifier.

    Check before if `s` is instance of string types.  See
    `safe_isidentifier`:func: for what "validity" means.

    '''
    from xoutil.eight import string_types
    return isinstance(s, string_types) and isfullidentifier(s)


def check_identifier(s):
    '''Check if `s` is a valid identifier.'''
    from xoutil.eight.string import isidentifier
    res = isidentifier(s)
    if res:
        return res
    else:
        msg = 'invalid identifier "{}"'
        raise TypeError(msg.format(s))



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.crypto
# ---------------------------------------------------------------------
# Copyright (c) 2014-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# Author: Medardo Rodriguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2014-04-22

'''General security tools.

Adds the ability to generate new passwords using a source pass-phrase and a
secury strong level.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)


__all__ = ('PASS_PHRASE_LEVEL_BASIC',
           'PASS_PHRASE_LEVEL_MAPPED',
           'PASS_PHRASE_LEVEL_MAPPED_MIXED',
           'PASS_PHRASE_LEVEL_MAPPED_DATED',
           'PASS_PHRASE_LEVEL_STRICT',
           'generate_password')


#: The most basic level (less ) for the password generation.
PASS_PHRASE_LEVEL_BASIC = 0

#: A level for simply mapping of several chars.
PASS_PHRASE_LEVEL_MAPPED = 1

#: Another "stronger" mapping level.
PASS_PHRASE_LEVEL_MAPPED_MIXED = 2

#: Appends the year after mapping.
PASS_PHRASE_LEVEL_MAPPED_DATED = 3

#: Totally scramble the result, making very hard to predict the result.
PASS_PHRASE_LEVEL_STRICT = 4

#: The default level for `generate_password`:func:
DEFAULT_PASS_PHRASE_LEVEL = PASS_PHRASE_LEVEL_MAPPED_DATED


#: A mapping from names to standards levels.  You may use these strings as
#: arguments for `level` in `generate_password`:func:.
PASS_LEVEL_NAME_MAPPING = {
    'basic': PASS_PHRASE_LEVEL_BASIC,
    'mapped': PASS_PHRASE_LEVEL_MAPPED,
    'mixed': PASS_PHRASE_LEVEL_MAPPED_MIXED,
    'dated': PASS_PHRASE_LEVEL_MAPPED_DATED,
    'random': PASS_PHRASE_LEVEL_STRICT
}


BASIC_PASSWORD_SIZE = 4    # bytes

#: An upper limit for generated password length.
MAX_PASSWORD_SIZE = 512


SAMPLE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"


def _normalize_level(level):
    '''Normalize the `level` argument.

    If passed a string, it must be a key in `PASS_LEVEL_NAME_MAPPING`:obj:.
    Otherwise it must be a valid level number.

    '''
    from xoutil.eight import string_types
    if isinstance(level, string_types):
        return PASS_LEVEL_NAME_MAPPING[level]
    else:
        return level


def generate_password(pass_phrase, level=DEFAULT_PASS_PHRASE_LEVEL):
    '''Generate a password from a source `pass-phrase` and a security `level`.

    :param pass_phrase: String pass-phrase to be used as base of password
                        generation process.

    :param level: Numerical security level (the bigger the more secure, but
                  don't exaggerate!).

    When `pass_phrase` is a valid string, `level` means a generation method.
    Each level implies all other with an inferior numerical value.

    There are several definitions with numerical values for `level` (0-4):

    `PASS_PHRASE_LEVEL_BASIC`:data:

        Generate the same pass-phrase, just removing invalid characters and
        converting the result to lower-case.

    `PASS_PHRASE_LEVEL_MAPPED`:data:

        Replace some characters with new values: ``'e'->'3'``, ``'i'->'1'``,
        ``'o'->'0'``, ``'s'->'5'``.

    `PASS_PHRASE_LEVEL_MAPPED_MIXED`:data:

        Consonants characters before 'M' (included) are converted to
        upper-case, all other are kept lower-case.

    `PASS_PHRASE_LEVEL_MAPPED_DATED`:data:

        Adds a suffix with the year of current date ("<YYYY>").

    `PASS_PHRASE_LEVEL_STRICT`:data:

        Randomly scramble previous result until unbreakable strong password is
        obtained.

    If `pass_phrase` is ``None`` or an empty string, generate a "secure salt"
    (a password not based in a source pass-phrase).  A "secure salt" is
    generated by scrambling the concatenation of a random phrases from the
    alphanumeric vocabulary.

    Returned password size is ``4*level`` except when a `pass-phrase` is given
    for `level` <= 4 where depend on the count of valid characters of
    `pass-phrase` argument, although minimum required is warranted.  When
    `pass-phrase` is ``None`` for `level` zero or negative, size ``4`` is
    assumed.  First four levels are considered weak.

    Maximum size is defined in the :data:`MAX_PASSWORD_SIZE` constant.

    Default level is :data:`PASS_PHRASE_LEVEL_MAPPED_DATED` when using a
    pass-phrase.

    '''
    from random import sample, randint
    from xoutil.string import slugify
    level = _normalize_level(level)
    size = MAX_PASSWORD_SIZE + 1    # means, return all calculated
    required = min(max(level, 1)*BASIC_PASSWORD_SIZE, MAX_PASSWORD_SIZE)
    if pass_phrase:
        # PASS_PHRASE_LEVEL_BASIC
        res = slugify(pass_phrase, '', invalid_chars='_')
        if level >= PASS_PHRASE_LEVEL_MAPPED:
            for (old, new) in ('e3', 'i1', 'o0', 's5'):
                res = res.replace(old, new)
        if level >= PASS_PHRASE_LEVEL_MAPPED_MIXED:
            for new in "BCDFGHJKLM":
                old = new.lower()
                res = res.replace(old, new)
        if level >= PASS_PHRASE_LEVEL_MAPPED_DATED:
            from datetime import datetime
            today = datetime.today()
            res += today.strftime("%Y")
        if level >= PASS_PHRASE_LEVEL_STRICT:
            size = required
    else:
        size = required
        count = randint(BASIC_PASSWORD_SIZE, 2*BASIC_PASSWORD_SIZE)
        res = ''.join(sample(SAMPLE, count))
    if size <= MAX_PASSWORD_SIZE:
        if len(res) < size:
            needed = (size - len(res)) // BASIC_PASSWORD_SIZE + 1
            res += generate_password(None, needed)
        res = ''.join(sample(res, size))
    return res[:size]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# Copyright (c) 2012 Manuel Vazquez

# TODO:  Take this out of xoutil.
from multiprocessing import Pool, cpu_count
from itertools import count, islice, repeat

try:
    from itertools import izip as zip
except ImportError:
    pass


def gcd(a, b):
    while a % b != 0:
        a, b = b, a % b
    return b


def job(args):
    which, show = args
    try:
        result = gcd(which, which + 2 ** 37 - 73)
        if show and result != 1:
            print(which, result)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    from datetime import datetime
    import argparse

    cpus = max(cpu_count() - 1, 1)
    parser = argparse.ArgumentParser('Run several workers calculating GCD '
                                     'for large numbers. Uses multiprocessing '
                                     'to test several CPUs.')
    parser.add_argument('--show', help='Whether to print or not the gcds',
                        action='store_true', dest='show', default=False)
    parser.add_argument('--workers',
                        help='Number of workers. Defaults to %d.' % cpus,
                        type=int, action='store', dest='workers',
                        default=cpus)
    parser.add_argument('--size',
                        help='The magnitude of the test. Represents how many '
                             'cycles are processed, exactly 10**MAGNITUDE; '
                             'so be nice!', type=int, action='store',
                        dest='magnitude', default=6)
    args = parser.parse_args()

    pool = Pool(processes=args.workers)
    start = datetime.now()
    try:
        print('Working...')
        pool.imap(job, zip(islice(count(2 ** 1028 + 1), 10 ** args.magnitude),
                           repeat(args.show)), chunksize=1024)
        pool.close()
        pool.join()
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
    end = datetime.now()
    total_seconds = (end - start).total_seconds()
    print('All workers (%d) finished in %f seconds' %
          (args.workers, total_seconds))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight._meta3
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2013-04-29

'''Implements the meta() function using the Py3k syntax.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

from . import _py3
assert _py3, 'This module should be loaded in Py3k only'


def metaclass(meta, **kwargs):
    from ._meta import Mixin

    if isinstance(meta, type) and issubclass(meta, type) and meta is not type:
        metabase = meta.__base__
    else:
        metabase = type

    class inner_meta(metabase):
        @classmethod
        def __prepare__(cls, name, bases, **kwargs):
            real = name != '__inner__'
            prepare = getattr(meta, '__prepare__', None) if real else None
            return prepare(name, bases, **kwargs) if prepare else dict()

        def __new__(cls, name, bases, attrs, **kw):
            if name != '__inner__':
                bases = tuple(b for b in bases if Mixin not in b.__bases__)
                return meta(name, bases, attrs)
            else:
                return type.__new__(cls, name, bases, attrs)

        def __init__(self, name, bases, attrs, **kw):
            pass

    from ._types import new_class
    kwds = dict(kwargs, metaclass=inner_meta)
    return new_class('__inner__', (Mixin, ), kwds=kwds)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.keywords
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-11-17

'''Tools for manage Python keywords as names.

Reserved Python keywords can't be used as attribute names, so this module
functions use the convention of rename the name using an underscore as
suffix when a reserved keyword is used as name.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

# TODO: Integrate -in one- this module with 'xoutil.eight.string' and rename
# it.


def suffix_kwd(name):
    '''Add an underscore suffix if name if a Python keyword.'''
    from keyword import iskeyword
    return '{}_'.format(name) if iskeyword(name) else name


def org_kwd(name):
    '''Remove the underscore suffix if name starts with a Python keyword.'''
    from keyword import iskeyword
    if name.endswith('_'):
        res = name[:-1]
        return res if iskeyword(res) else name
    else:
        return name


def getkwd(obj, name, default=None):
    '''Like `getattr` but taking into account Python keywords.'''
    return getattr(obj, suffix_kwd(name), default)


def setkwd(obj, name, value):
    '''Like `setattr` but taking into account Python keywords.'''
    setattr(obj, suffix_kwd(name), value)


def delkwd(obj, name):
    '''Like `delattr` but taking into account Python keywords.'''
    delattr(obj, suffix_kwd(name))


def kwd_getter(obj):
    '''partial(getkwd, obj)'''
    from functools import partial
    return partial(getkwd, obj)


def kwd_setter(obj):
    '''partial(setkwd, obj)'''
    from functools import partial
    return partial(setkwd, obj)


def kwd_deleter(obj):
    '''partial(delkwd, obj)'''
    from functools import partial
    return partial(delkwd, obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.versions
# ---------------------------------------------------------------------
# Copyright (c) 2016-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-04-28

'''Versions API

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

from xoutil.decorator import singleton

try:
    _str_types = (str, unicode)
except:
    _str_types = (str,)


def _safe_int(arg):
    try:
        if isinstance(arg, _str_types):
            return int(arg)
        else:
            return arg
    except ValueError:
        return arg


def _crucial_parts(info):
    '''Count the important parts (heading integers) of a version info'''
    i, ok = 0, True
    while i < len(info) and ok:
        v = info[i]
        if isinstance(v, int):
            i += 1
        else:
            ok = False
    return i


def _check(info, head=False):
    '''Check a version info.

    `info` could be a string or any iterable.  if `hard`, the 3 required
    integer components must be supplied at the beginning.

    If `head` is true, return only the crucial parts.

    '''
    import re
    from collections import Mapping, Set
    MAX_IDX = 3
    if isinstance(info, int):
        res = (info, )
    elif isinstance(info, float):
        res = tuple(int(i) for i in str(info).split('.'))
    elif isinstance(info, _str_types):
        try:
            res = (int(info), )
        except ValueError:
            regex = re.compile('^(\d+)(?:[.](\d+))?(?:[.](\d+))?')
            m = regex.match(info)
            if m:
                res = tuple(int(g) for g in m.groups() if g is not None)
            else:
                raise TypeError('version info check got an invalid string')
    else:
        # a valid iterable
        if not isinstance(info, (Mapping, Set)):
            try:
                res = tuple(_safe_int(c) for c in info)
            except TypeError:
                res = None
        else:
            res = None
        if res is None:
            # TODO: from xoutil.eight import type_name
            msg = 'version info check got an invalid value of type "{}"'
            raise TypeError(msg.format(type(info).__name__))
    count = _crucial_parts(res)
    if 1 <= count <= MAX_IDX:
        return res[:count] if head else res
    else:
        msg = ('a version info need at least one and at most {} heading '
               'integer components; got "{}"')
        raise TypeError(msg.format(MAX_IDX, count))


class ThreeNumbersVersion(tuple):
    '''A more structured version info for packages.

    This class can't be instanced, each sub-class must be treated as
    singleton.  Must be a tuple with three first components as integers
    'major', 'minor', and 'micro'.  Any custom implementation could have extra
    components.

    Instances of this class can be compared with a variety of value types:

    - An integer with the 'major' component.

    - A float with the ('major', 'minor') components.

    - A string is converted to a version tuple before compare it.

    But comparison only is relevant for heading integers.

    .. versionadded:: 1.8.0

    '''
    def __new__(cls, info):
        return super(ThreeNumbersVersion, cls).__new__(cls, _check(info))

    @property
    def major(self):
        return self[0]

    @property
    def minor(self):
        return self[1]

    @property
    def micro(self):
        return self[2]

    def to_float(self):
        return float('{}.{}'.format(*self[:2]))

    def __eq__(self, other):
        aux = _check(other, head=True)
        this = self[:len(aux)]
        return this == aux

    def __lt__(self, other):
        aux = _check(other, head=True)
        count = _crucial_parts(self)
        this = self[:count]
        return this < aux

    def __gt__(self, other):
        aux = _check(other, head=True)
        count = _crucial_parts(self)
        this = self[:count]
        return this > aux

    def __ne__(self, other):
        return not (self == other)

    def __le__(self, other):
        return not (self > other)

    def __ge__(self, other):
        return not (self < other)


@singleton
class python_version(ThreeNumbersVersion):
    '''Current Python version.

    Initialized with `~sys.version_info`:obj: 5 components tuple.

    Extra components (besides 'major', 'minor', and 'micro') are:
    'releaselevel' (a string that could be 'alpha', 'beta', 'candidate', or
    'final'), and 'serial' (an integer).  The attribute 'pypy' could be used
    to determine if this is a PyPy instance or not.

    '''
    def __new__(cls):
        import sys
        self = ThreeNumbersVersion.__new__(cls, sys.version_info)
        self.pypy = sys.version.find('PyPy') >= 0
        return self

    @property
    def releaselevel(self):
        return self[3]

    @property
    def serial(self):
        return self[4]


def _get_mod_version(mod):
    '''Get a valid version from a module.

    Used internally by `PackageVersion`:class:.

    '''
    valid_names = ('VERSION_INFO', 'VERSION', 'version_info', 'version',
                   '__VERSION__', '__version__')
    i = 0
    res = None
    while res is None and i < len(valid_names):
        name = valid_names[i]
        version = getattr(mod, name, None)
        if version is not None:
            try:
                res = _check(version)
            except:
                pass
        i += 1
    if not res:
        try:
            import os
            path = os.path.dirname(os.__file__)
            if mod.__file__.startswith(path):
                res = python_version
        except:
            pass
    return res


class PackageVersion(ThreeNumbersVersion):
    '''Current Package version.

    Extra components (besides 'major', 'minor', and 'micro') are:
    'releaselevel' (a string that could be 'alpha', 'beta', 'candidate', or
    'final'), and 'serial' (an integer).  The attribute 'pypy' could be used
    to determine if this is a PyPy instance or not.

    '''
    def __new__(cls, package_name):
        info = cls._find_version(package_name)
        if info:
            return super(PackageVersion, cls).__new__(cls, info)
        else:
            msg = '{}() could not determine a valid version'
            raise TypeError(msg.format(cls.__name__))

    @staticmethod
    def _find_version(package_name):
        if package_name in ('__builtin__', 'builtins'):
            return python_version
        else:
            res = None
            while not res and package_name:
                try:
                    import pkg_resources
                    dist = pkg_resources.get_distribution(package_name)
                    try:
                        res = dist.parsed_version.base_version
                    except:
                        res = dist.version
                except:
                    from importlib import import_module
                    try:
                        mod = import_module('.'.join((package_name, 'release')))
                        res = _get_mod_version(mod)
                    except:
                        try:
                            mod = import_module(package_name)
                            res = _get_mod_version(mod)
                        except:
                            mod = __import__(package_name)
                            res = _get_mod_version(mod)
                if not res:
                    aux = package_name.rsplit('.', 1)
                    package_name = aux[0] if len(aux) > 1 else ''
            return res



# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.future.textwrap
# ---------------------------------------------------------------------
# Copyright (c) 2014-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the Python Software Licence as of Python 3.3.
#
# Created on 2014-02-26
# Migrated to 'future' on 2016-09-18

'''Text wrapping and filling.'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from textwrap import *    # noqa
import textwrap as _stdlib
from textwrap import __all__    # noqa
__all__ = list(__all__)

from xoutil.deprecation import deprecate_linked
deprecate_linked()
del deprecate_linked


def dedent(text, skip_firstline=False):
    r'''Remove any common leading whitespace from every line in text.

    This can be used to make triple-quoted strings line up with the left edge
    of the display, while still presenting them in the source code in indented
    form.

    Note that tabs and spaces are both treated as whitespace, but they are not
    equal: the lines ``"    hello"`` and ``"\thello"`` are considered to have
    no common leading whitespace.

    If `skip_firstline` is True, the first line is separated from the rest of
    the body.  This helps with docstrings that follow `257`:pep:.

    .. warning:: The `skip_firstline` argument is missing in standard library.

    '''
    if skip_firstline:
        parts = text.split('\n', 1)
        if len(parts) > 1:
            subject, body = parts
        else:
            subject, body = parts[0], ''
        result = _stdlib.dedent(subject)
        if body:
            result += '\n' + _stdlib.dedent(body)
    else:
        result = _stdlib.dedent(text)
    return result


try:
    indent    # noqa
except NameError:
    # The following is Copyright (c) of the Python Software Foundation.
    def indent(text, prefix, predicate=None):
        """Adds 'prefix' to the beginning of selected lines in 'text'.

        If 'predicate' is provided, 'prefix' will only be added to the lines
        where 'predicate(line)' is True. If 'predicate' is not provided, it
        will default to adding 'prefix' to all non-empty lines that do not
        consist solely of whitespace characters.

        .. note:: Backported from Python 3.3.  In Python 3.3 this is an alias.

        """
        if predicate is None:
            def predicate(line):
                return line.strip()

        def prefixed_lines():
            for line in text.splitlines(True):
                yield (prefix + line if predicate(line) else line)
        return ''.join(prefixed_lines())

    __all__.append('indent')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight._meta
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-10-30

'''Private module for all mix-ins definitions.

This module is private because is used for two client modules for the `mixin`
concept.  See `Mixin`:class: for more information.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


class Mixin(object):
    '''Base-class to several mix-in kinds.

    - Syntax unification for meta-classes between Python 2 and 3.  See
      `xoutil.eight.meta.metaclass`:func: for more information.  For example::

        >>> from xoutil.eight.abc import ABCMeta
        >>> from xoutil.eight.meta import metaclass
        >>> class Foobar(metaclass(ABCMeta)):
        ...     pass
        >>> (Foobar.__bases__ == (object, ), type(Foobar) is ABCMeta)
        (True, True)

    - Helper base classes that when used as a base force a meta-class but
      nothing else (the new class don't have helper-class in its bases).  See
      `xoutil.eight.mixins.helper_class`:func: for more information.  A good
      example is our example of `xoutil.eight.abc.ABC`, if used as a base,
      `ABCMeta` is imposed::

        >>> from xoutil.eight.abc import ABC, ABCMeta
        >>> class Foobar(ABC):
        ...     pass
        >>> (Foobar.__bases__ == (object, ), type(Foobar) is ABCMeta)
        (True, True)

    - Class templates that can be used later to weave main classes composed
      for those mix-ins.  For example::

        >>> from xoutil.eight.abc import ABCMeta
        >>> from xoutil.eight.mixins import mixin
        >>> class Foo(object):
        ...     pass
        >>> class Bar(dict):
        ...     pass
        >>> class EggMeta(type):
        ...     pass
        >>> Test = mixin(Foo, Bar, meta=(EggMeta, ABCMeta), name='Test')

    - Weaver a main-class using several parts.  For example::

        >>> class Foobar(list, mixin(Test)):
        ...     pass

    '''



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.meta
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2015-02-25

'''Implements the metaclass() function using the Py3k syntax.

'''

from . import _py3
from ._meta import Mixin    # noqa

if _py3:
    from ._meta3 import metaclass
else:
    from ._meta2 import metaclass


metaclass.__doc__ = '''Define the metaclass of a class.

    .. versionadded:: 1.7.0  It's available as
       `xoutil.objects.metaclass`:func: since 1.4.1.  That alias is now
       deprecated and will be removed in 1.8.

    This function allows to define the metaclass of a class equally in Python
    2 and 3.

    Usage::

     >>> class Meta(type):
     ...   pass

     >>> class Foobar(metaclass(Meta)):
     ...   pass

     >>> class Spam(metaclass(Meta), dict):
     ...   pass

     >>> type(Spam) is Meta
     True

     >>> Spam.__bases__ == (dict, )
     True

    .. versionadded:: 1.5.5 The `kwargs` keywords arguments with support for
       ``__prepare__``.

    Metaclasses are allowed to have a ``__prepare__`` classmethod to return
    the namespace into which the body of the class should be evaluated.  See
    `3115`:pep:.

    .. warning:: The `3115`:pep: is not possible to implement in Python 2.7.

       Despite our best efforts to have a truly compatible way of creating
       meta classes in both Python 2.7 and 3.0+, there is an inescapable
       difference in Python 2.7.  The `3115`:pep: states that ``__prepare__``
       should be called before evaluating the body of the class.  This is not
       possible in Python 2.7, since ``__new__`` already receives the
       attributes collected in the body of the class.  So it's always too late
       to call ``__prepare__`` at this point and the Python 2.7 interpreter
       does not call it.

       Our approach for Python 2.7 is calling it inside the ``__new__`` of a
       "side" metaclass that is used for the base class returned.  This means
       that ``__prepare__`` is called **only** for classes that use the
       `metaclass`:func: directly.  In the following hierarchy::

         class Meta(type):
              @classmethod
              def __prepare__(cls, name, bases, **kwargs):
                  from xoutil.future.collections import OrderedDict
                  return OrderedDict()

         class Foo(metaclass(Meta)):
              pass

         class Bar(Foo):
              pass

       when creating the class ``Bar`` the ``__prepare__()`` class method is
       not called in Python 2.7!

    .. seealso:: `xoutil.future.types.prepare_class`:func: and
       `xoutil.future.types.new_class`:func:.

    .. warning::

       You should always place your metaclass declaration *first* in the list
       of bases. Doing otherwise triggers *twice* the metaclass' constructors
       in Python 3.1 or less.

       If your metaclass has some non-idempotent side-effect (such as
       registration of classes), then this would lead to unwanted double
       registration of the class::

        >>> class BaseMeta(type):
        ...     classes = []
        ...     def __new__(cls, name, bases, attrs):
        ...         res = super(BaseMeta, cls).__new__(cls, name, bases, attrs)
        ...         cls.classes.append(res)   # <-- side effect
        ...         return res

        >>> class Base(metaclass(BaseMeta)):
        ...     pass

        >>> class SubType(BaseMeta):
        ...     pass

        >>> class Egg(metaclass(SubType), Base):   # <-- metaclass first
        ...     pass

        >>> Egg.__base__ is Base   # <-- but the base is Base
        True

        >>> len(BaseMeta.classes) == 2
        True

        >>> class Spam(Base, metaclass(SubType)):
        ...     'Like "Egg" but it will be registered twice in Python 2.x.'

       In this case the registration of Spam ocurred twice::

        >>> BaseMeta.classes  # doctest: +SKIP
        [<class Base>, <class Egg>, <class Spam>, <class Spam>]

       Bases, however, are just fine::

        >>> Spam.__bases__ == (Base, )
        True

    .. versionadded:: 1.7.1 Now are accepted atypical meta-classes, for
       example functions or any callable with the same arguments as those that
       type accepts (class name, tuple of base classes, attributes mapping).

'''

metaclass.__module__ = __name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.decorator.meta
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2013-01-03
#

'''Decorator-making facilities.

This module provides a signature-keeping version of the
`xoutil.decorators.decorator`:func:, which is now deprecated in favor of this
module's version.

We scinded the decorator-making facilities from decorators per se to allow the
module `xoutil.deprecation`:mod: to be used by decorators and at the same
time, implement the decorator `~xoutil.deprecation.deprecated`:func: more
easily.


This module is an adapted work from the decorator version 3.3.2 package and is
copyright of its owner as stated below. Adaptation work is done by Merchise.

Original copyright and license notices from decorator package:

    Copyright (c) 2005-2011, Michele Simionato

    All rights reserved.

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.  Redistributions in
    bytecode form must reproduce the above copyright notice, this list of
    conditions and the following disclaimer in the documentation and/or other
    materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

import sys
import re
import inspect

from functools import wraps, partial
from types import FunctionType as function

if sys.version_info[0] == 3:
    from inspect import getfullargspec as _getfullargspec
else:
    from inspect import getargspec as _getfullargspec


__all__ = ('FunctionMaker', 'flat_decorator', 'decorator')


DEF = re.compile('\s*def\s*([_\w][_\w\d]*)\s*\(')


# basic functionality
class FunctionMaker(object):
    """
    An object with the ability to create functions with a given signature.
    It has attributes name, doc, module, signature, defaults, dict and
    methods update and make.
    """
    def __init__(self, func=None, name=None, signature=None,
                 defaults=None, doc=None, module=None, funcdict=None):
        self.shortsignature = signature
        if func:
            # func can be a class or a callable, but not an instance method
            self.name = func.__name__
            if self.name == '<lambda>':  # small hack for lambda functions
                self.name = '_lambda_'
            self.doc = func.__doc__
            self.module = func.__module__
            if inspect.isfunction(func):
                argspec = _getfullargspec(func)
                for a in ('args', 'varargs', 'varkw', 'defaults', 'kwonlyargs',
                          'kwonlydefaults', 'annotations'):
                    setattr(self, a, getattr(argspec, a, None))
                for i, arg in enumerate(self.args):
                    setattr(self, 'arg%d' % i, arg)
                self.signature = inspect.formatargspec(
                    formatvalue=lambda val: "", *argspec)[1:-1]
                allargs = list(self.args)
                if self.varargs:
                    allargs.append('*' + self.varargs)
                if self.varkw:
                    allargs.append('**' + self.varkw)
                try:
                    self.shortsignature = ', '.join(allargs)
                except TypeError:
                    # exotic signature, valid only in Python 2.X
                    self.shortsignature = self.signature
                self.dict = func.__dict__.copy()
        # func=None happens when decorating a caller
        if name:
            self.name = name
        if signature is not None:
            self.signature = signature
        if defaults:
            self.defaults = defaults
        if doc:
            self.doc = doc
        if module:
            self.module = module
        if funcdict:
            self.dict = funcdict
        # check existence required attributes
        assert hasattr(self, 'name')
        if not hasattr(self, 'signature'):
            raise TypeError('You are decorating a non function: %s' % func)

    def update(self, func, **kw):
        "Update the signature of func with the data in self"
        func.__name__ = self.name
        func.__doc__ = getattr(self, 'doc', None)
        func.__dict__ = getattr(self, 'dict', {})
        func.func_defaults = getattr(self, 'defaults', ())
        func.__kwdefaults__ = getattr(self, 'kwonlydefaults', None)
        callermodule = sys._getframe(3).f_globals.get('__name__', '?')
        func.__module__ = getattr(self, 'module', callermodule)
        func.__dict__.update(kw)

    def make(self, src_templ, evaldict=None, addsource=False, **attrs):
        "Make a new function from a given template and update the signature"
        src = src_templ % vars(self)  # expand name and signature
        evaldict = evaldict or {}
        mo = DEF.match(src)
        if mo is None:
            raise SyntaxError('not a valid function template\n%s' % src)
        name = mo.group(1)  # extract the function name
        names = set([name] + [arg.strip(' *') for arg in
                              self.shortsignature.split(',')])
        for n in names:
            if n in ('_func_', '_call_'):
                raise NameError('%s is overridden in\n%s' % (n, src))
        if not src.endswith('\n'):  # add a newline just for safety
            src += '\n'   # this is needed in old versions of Python
        try:
            code = compile(src, '<string>', 'single')
            # print >> sys.stderr, 'Compiling %s' % src
            eval(code, evaldict, evaldict)
        except:
            print >> sys.stderr, 'Error in generated code:'
            print >> sys.stderr, src
            raise
        func = evaldict[name]
        if addsource:
            attrs['__source__'] = src
        self.update(func, **attrs)
        return func

    @classmethod
    def create(cls, obj, body, evaldict, defaults=None,
               doc=None, module=None, addsource=True, **attrs):
        """
        Create a function from the strings name, signature and body.
        "evaldict" is the evaluation dictionary. If addsource is true an
        attribute __source__ is added to the result. The attributes attrs are
        added,
        if any.
        """
        try:
            string_types = (str, unicode)
        except NameError:
            string_types = (str,)
        if isinstance(obj, string_types):  # "name(signature)"
            obj = str(obj)
            name, rest = obj.strip().split(str('('), 1)
            signature = rest[:-1]   # strip a right parens
            func = None
        else:   # a function
            name = None
            signature = None
            func = obj
        self = cls(func, name, signature, defaults, doc, module)
        ibody = '\n'.join('    ' + line for line in body.splitlines())
        return self.make('def %(name)s(%(signature)s):\n' + ibody,
                         evaldict, addsource, **attrs)


def flat_decorator(caller, func=None):
    """
    Creates a signature keeping decorator.

    ``decorator(caller)`` converts a caller function into a decorator.

    ``decorator(caller, func)`` decorates a function using a caller.
    """
    if func is not None:    # returns a decorated function
        evaldict = func.func_globals.copy()
        evaldict['_call_'] = caller
        evaldict['_func_'] = func
        return FunctionMaker.create(
            func, "return _call_(_func_, %(shortsignature)s)",
            evaldict, undecorated=func, __wrapped__=func)
    else:   # returns a decorator
        if isinstance(caller, partial):
            return partial(decorator, caller)
        # otherwise assume caller is a function
        try:
            first = inspect.getargspec(caller)[0][0]    # first arg
            deco_sign = '%s(%s)' % (caller.__name__, first)
            deco_body = 'return flat_decorator(_call_, %s)' % first
        except IndexError:
            deco_sign = '%s()' % caller.__name__
            deco_body = 'return _call_'
        evaldict = caller.func_globals.copy()
        evaldict['_call_'] = caller
        evaldict['flat_decorator'] = evaldict['decorator'] = flat_decorator
        return FunctionMaker.create(
            deco_sign,
            deco_body,
            evaldict, undecorated=caller, __wrapped__=caller,
            doc=caller.__doc__, module=caller.__module__)
# -- End of decorators package


# FIX: This meta-decorator fails in some scenarios (old classes?)
def decorator(caller):
    '''Eases the creation of decorators with arguments.  Normally a decorator
    with arguments needs three nested functions like this::

        def decorator(*decorator_arguments):
            def real_decorator(target):
                def inner(*args, **kwargs):
                    return target(*args, **kwargs)
                return inner
            return real_decorator

    This decorator reduces the need of the first level by comprising both into
    a single function definition. However it does not removes the need for an
    ``inner`` function::

        >>> @decorator
        ... def plus(target, value):
        ...    from functools import wraps
        ...    @wraps(target)
        ...    def inner(*args):
        ...        return target(*args) + value
        ...    return inner

        >>> @plus(10)
        ... def ident(val):
        ...     return val

        >>> ident(1)
        11

    A decorator with default values for all its arguments (except, of course,
    the first one which is the decorated `target`) may be invoked
    without parenthesis::

        >>> @decorator
        ... def plus2(func, value=1, missing=2):
        ...    from functools import wraps
        ...    @wraps(func)
        ...    def inner(*args):
        ...        print(missing)
        ...        return func(*args) + value
        ...    return inner

        >>> @plus2
        ... def ident2(val):
        ...     return val

        >>> ident2(10)
        2
        11

    But (if you like) you may place the parenthesis::

        >>> @plus2()
        ... def ident3(val):
        ...     return val

        >>> ident3(10)
        2
        11

    However, this is not for free, you cannot pass a single positional argument
    which type is a function::

        >>> def p():
        ...    print('This is p!!!')

        >>> @plus2(p)   # doctest: +ELLIPSIS
        ... def dummy():
        ...    print('This is dummy')
        Traceback (most recent call last):
            ...
        TypeError: p() takes ...

    The workaround for this case is to use a keyword argument.
    '''
    @wraps(caller)
    def outer_decorator(*args, **kwargs):
        try:
            from zope.interface import Interface
        except ImportError:
            Interface = None
            # from xoutil.symbols import Unset as Interface
        if (len(args) == 1 and not kwargs and
            (isinstance(args[0], (function, type)) or
             issubclass(type(args[0]), type(Interface)))):
            # This tries to solve the case of missing () on the decorator::
            #
            #    @decorator
            #    def somedec(func, *args, **kwargs)
            #        ...
            #
            #    @somedec
            #    def decorated(*args, **kwargs):
            #        pass
            #
            # Notice, however, that this is not general enough, since we try
            # to avoid inspecting the calling frame to see if the () are in
            # place.
            func = args[0]
            return caller(func)
        elif len(args) > 0 or len(kwargs) > 0:
            def _decorator(func):
                return partial(caller, **kwargs)(*((func, ) + args))
            return _decorator
        else:
            return caller
    return outer_decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.dim.currencies
# ---------------------------------------------------------------------
# Copyright (c) 2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-09-14

'''Concrete numbers for money.

You may have 10 dollars and 5 euros in your wallet, that does not mean that
you have 15 of anything (but bills, perhaps).  Though you may *evaluate* your
cash in any other currency you don't have that value until you perform an
exchange with a given rate.

This module support the family of currencies.  Usage::

  >>> from xoutil.dim.currencies import Rate, Valuation, currency
  >>> dollar = USD = currency('USD')
  >>> euro = EUR = currency('EUR')
  >>> rate = 1.19196 * USD/EUR

  >>> isinstance(dollar, Valuation)
  True

  >>> isinstance(rate, Rate)
  True

  # Even 0 dollars are a valuation
  >>> isinstance(dollar - dollar, Valuation)
  True

  # But 1 is not a value nor a rate
  >>> isinstance(dollar/dollar, Valuation) or isinstance(dollar/dollar, Rate)
  False


Currency names are case-insensitive.  We don't check the currency name is
listed in `ISO 4217`_.  So currency ``MVA`` is totally acceptable in this
module.

We don't download rates from any source.

This module allows you to trust your computations of money by allowing only
sensible operations::

  >>> dollar + euro  # doctest: +ELLIPSIS
  Traceback (...)
  ...
  OperandTypeError: unsupported operand type(s) for +: '{USD}/{}' and '{EUR}/{}


If you convert your euros to dollars::

  >>> dollar + rate * euro
  2.19196::{USD}/{}

  # Or your dollars to euros
  >>> dollar/rate + euro
  1.83895432733::{EUR}/{}


.. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.eight.meta import metaclass


class ValueType(type):
    def __instancecheck__(self, which):
        from .meta import Quantity
        if isinstance(which, Quantity):
            return any(
                which.signature is currency.signature
                for currency in _Currency.units.values()
            )
        else:
            return False


class Valuation(metaclass(ValueType)):
    pass


class RateType(type):
    def __instancecheck__(self, which):
        from .meta import Quantity
        if isinstance(which, Quantity):
            top, bottom = which.signature.top, which.signature.bottom
            if len(top) == len(bottom) == 1:
                iscurrency = lambda s: isinstance(s[0], _Currency)
                return iscurrency(top) and iscurrency(bottom)
            else:
                return False
        else:
            return False


class Rate(metaclass(RateType)):
    pass


class _Currency(object):
    instances = {}
    units = {}

    def __new__(cls, name):
        from .meta import Quantity, Signature
        name = name.upper()
        res = cls.instances.get(name, None)
        if res is None:
            res = super(_Currency, cls).__new__(cls)
            res.name = name
            cls.instances[name] = res
            cls.units[name] = Quantity(1, Signature(top=(res, )))
        return res

    def __str__(self):
        return self.name

    @property
    def unit(self):
        return self.units[self.name]


def currency(name):
    '''Get the canonical value for the given currency `name`.

    '''
    return _Currency(name).unit
